# Copyright 2015 Jeff Vierstra

import numpy as np

class genomic_interval(object):
	"""Class that implements BED-style object
	"""

	def __init__(self, chrom, start, end, name = '.', score = None, strand = None):
		self.chrom = str(chrom)
		self.start = int(start)
		self.end = int(end)
		self.name = str(name)
		self.score = score
		self.strand = strand
		
	def __len__(self):
		"""returns length of element
		"""

		return self.end - self.start
	
	def __str__(self):
		"""returns a string-formated version of the element for printing
		"""
		return '\t'.join( [ str(x) for x in [self.chrom, self.start, self.end] ] )

	def widen(self, x, inplace=False):
		"""expands the coordinates

		:param x: number of nucletides to expand (can also be negative)
		:param inplace: modify estixing element or return a modified copy
		:type x: int
		:type inplace: bool
		:rtype: genomic_interval
		"""
		
		if inplace:
			self.start-=x
			self.end+=x
			return self
		return genomic_interval(self.chrom, self.start-x, self.end+x, self.name, self.strand)

	def shift(self, x, inplace=False):
		if inplace:
			self.start+=x
			self.end+=x
			return self
		else:	
			return genomic_interval(self.chrom, self.start+x, self.end+x, self.name, self.stand)

class genomic_interval_set(object):

	def __init__(self, iterator = []):
		self.intervals = []
		for interval in iterator:
			self.intervals.append(interval)
	
	def __len__(self):
		return len(self.intervals)

	def __iter__(self):
		for x in self.intervals:
			yield x

	def __getitem__(self, i):
		return self.intervals[i]

	def __iadd__(self, other):
		if type(other) == genomic_interval_set:
			self.intervals.extend(other.intervals)
		else:
			self.intervals.append(other)
		return self

def bin_intervals(intervals, bins):
	"""Bin genomic intervals by score thresholds

	Args:
		intervals (genomic_interval_set): a set of genomic intervals
						that contain a score value
		bins (array): a list of values comprising desired bins
						(must by monotonically increasing or decreasing)
	Returns:
		(list): a list of indices for each input interval to a corresponding bin
	"""
	scores = np.zeros(len(intervals))
	for i in range(len(intervals)):
		scores[i] = intervals[i].score
	return np.digitize(scores, thresh)





__all__ = ["bed", "genomic_interval", "helpers"]

from .genomic_interval import *

import pysam

def load_data(filepath, interval, data_columns=[5], dtype=np.float):

	""" Loads numeric data columns from a BED-format TABIX file """

	res = np.zeros((len(interval), len(data_columns)), dtype = dtype)

	tabix = pysam.TabixFile(filepath)

	for row in tabix.fetch(interval.chrom, interval.start, interval.end, parser = pysam.asTuple()):
	    i = int(row[1])- interval.start
	    
	    for j, col in enumerate(data_columns):
	    	res[i, j] = dtype(row[col-1])

	tabix.close()
	    
	return res


def load_segmented_data(filepath, interval):
	
	res = genomic_interval_set()

	tabix = pysam.TabixFile(filepath)
	for row in tabix.fetch(interval.chrom, interval.start, interval.end, parser = pysam.asTuple()):
		
		chrom=row[0]
		start = int(row[1])
		end = int(row[2])
		
		try:
			name=row[3]
		except:
			name='.'

		try:
			score=float(row[4])
		except:
			score=-np.inf
	
		try:
			strand=row[5]
		except:
			strand='+'

		res += genomic_interval(chrom, start, end, name=name, score=score, strand=strand) 
	
	tabix.close()

	return res




# Copyright 2016 Jeff Vierstra

import numpy as np

from matplotlib.ticker import MaxNLocator
from matplotlib.transforms import blended_transform_factory

from .. import load_data
from .track import track, segment

class continuous_data_track(track):

	def __init__(self, interval, data=None, **kwargs):
		super(continuous_data_track, self).__init__(interval, **kwargs)
		self.data = data

	def load_data(self, filepath, column=5, dtype=float):
		self.data = load_data(filepath, interval, [column], dtype)[:,0]


	def density(self, vals, window_size = 150, step_size = 20):
		"""Smooth data in density windows"""
		
		pos = np.arange(0, len(vals)-window_size, step_size)
		s = np.zeros(len(pos))
		for i, j in enumerate(pos):
			s[i] = np.sum(vals[j:j+window_size])
		return pos + (window_size/2) + self.interval.start, s

	def simplify(self, step = 100):
		"""Simplify data points for fast plotting -- there is probably a better way of doing this..."""
		
		xx = []
		yy = []
		for i in range(step, len(self.data)-step, step):
			xx.append(self.interval.start + i)
			yy.append(np.amax(self.data[i-step:i+step]))
		return np.array(xx), np.array(yy)

	def step(self, vals, xaxis = False, step_interval = 0):
		"""Creates a step-style plot"""

		if xaxis and step_interval == 0:
			step_interval = abs(vals[1] - vals[0]) / 2.0
		step_vals = np.array(list(zip(vals - step_interval, vals + step_interval))).ravel()
		return step_vals

	def format_axis(self, ax):
		super(continuous_data_track, self).format_axis(ax)

		locator = MaxNLocator(3, prune = 'upper')
		ax.yaxis.set(visible = True, major_locator = locator)

		if 'min' in self.options:
			ax.set_ylim(bottom = self.options['min'])

		if 'max' in self.options:
			ax.set_ylim(top = self.options['max'])


	def render(self, ax):

		if self.data is None:
			raise Exception("No data loaded!")

		self.format_axis(ax)
		self.format_spines(ax, remove_spines=['top', 'right'])

		if 'density' in self.options:
			xx, yy = self.density(self.data, window_size = self.options['density']['window_size'])
			xs = self.step(xx, xaxis = True)
			ys = self.step(yy) / self.options['density']['norm_factor']
		else:
			xs = self.step(np.arange(self.interval.start, self.interval.end), xaxis = True)
			ys = self.step(self.data)


		if 'simplify' in self.options:
			xx, yy = self.simplify(step = self.options['simplify'])
			xs = self.step(xx, xaxis = True)
			ys = self.step(yy, xaxis = False)

		(ybot, ytop) = ax.get_ylim()
		ys[ys > ytop] = ytop
		ys[ys < ybot] = ybot

		if 'fill_between' in self.options:
			ax.fill_between(xs, np.ones(len(xs)) * self.options["fill_between"], ys, 
				facecolor=self.options["facecolor"], 
				edgecolor=self.options["edgecolor"],
				linewidth=self.options["lw"],
				linestyle=self.options["ls"])
		else:
			ax.plot(xs, y, 
				color=self.options["edgecolor"], 				
				lw=self.options["lw"],
				ls=self.options["ls"])

		# change to linecollection
		#if 'clip_bar' in self.options:
		#	trans = blended_transform_factory(ax.transData, ax.transAxes)
		#	for x0, x1 in segment(self.data, ytop):
		#		ax.plot([x0 + self.interval.start, x1 + self.interval.start], [1, 1], color = self.options['clip_bar']['color'], lw = self.options['clip_bar']['lw'], transform = trans, clip_on = False)

		super(continuous_data_track, self).render(ax)





from matplotlib.transforms import Bbox, TransformedBbox, blended_transform_factory
from mpl_toolkits.axes_grid1.inset_locator import BboxPatch, BboxConnector, BboxConnectorPatch

# Connect two bounding boxes in possibly different axes
def connect_bbox(bbox1, bbox2, loc1a, loc2a, loc1b, loc2b, prop_lines, prop_patches=None):
    if prop_patches is None:
        prop_patches = prop_lines.copy()
        prop_patches["alpha"] = prop_patches.get("alpha", 1)*0.2

    c1 = BboxConnector(bbox1, bbox2, loc1=loc1a, loc2=loc2a, **prop_lines)
    c1.set_clip_on(False)

    c2 = BboxConnector(bbox1, bbox2, loc1=loc1b, loc2=loc2b, **prop_lines)
    c2.set_clip_on(False)

    bbox1_patch= BboxPatch(bbox1, **prop_patches)
    bbox2_patch = BboxPatch(bbox2, **prop_patches)

    # connects two bbox-es with a quadrilateral
    p = BboxConnectorPatch(bbox1, bbox2, loc1a=loc1a, loc2a=loc2a, loc1b=loc1b, loc2b=loc2b, **prop_patches)
    p.set_fill(True)
    p.set_clip_on(False)

    return c1, c2, bbox1_patch, bbox2_patch, p

def zoom_effect(ax_zoomed, ax_origin, xlims = None, orientation='below', **kwargs):
    """
    ax_zoomed : zoomed axes
    ax_origin:  the main axes
    (xmin,xmax) : the limits of the colored area in both plot axes.

    connect ax1 & ax2. The x-range of (xmin, xmax) in both axes will
    be marked.  The keywords parameters will be used ti create
    patches.

    """
    if xlims is None:
        tt = ax_zoomed.transScale + (ax_zoomed.transLimits + ax_origin.transAxes)
        transform = blended_transform_factory(ax_origin.transData, tt)

        bbox_zoomed=ax_zoomed.bbox
        bbox_origin=TransformedBbox(ax_zoomed.viewLim, transform)
    else:
        transform_zoomed=blended_transform_factory(ax_zoomed.transData, ax_zoomed.transAxes)
        transform_origin=blended_transform_factory(ax_origin.transData, ax_origin.transAxes)
    
        bbox_zoomed=TransformedBbox(Bbox.from_extents(xlims[0], 0, xlims[1], 1), transform_zoomed)
        bbox_origin=TransformedBbox(Bbox.from_extents(xlims[0], 0, xlims[1], 1), transform_origin)

    prop_patches = kwargs.copy()
    prop_patches["ec"] = "none"
    prop_patches["alpha"] = 0.2

    if orientation=='below':
        loc1a=2
        loc2a=3
        loc1b=1
        loc2b=4
    elif orientation=='above':
        loc1a=3
        loc2a=2
        loc1b=4
        loc2b=1
    else:
        raise Exception("orientation '%s' not recognized" % orientation)

    c1, c2, bbox_zoomed_patch, bbox_origin_patch, p = \
        connect_bbox(bbox_zoomed, bbox_origin,
                     loc1a=loc1a, loc2a=loc2a, loc1b=loc1b, loc2b=loc2b,
                     prop_lines=kwargs, prop_patches=prop_patches)

    ax_zoomed.add_patch(bbox_zoomed_patch)
    ax_origin.add_patch(bbox_origin_patch)
    ax_origin.add_patch(c1)
    ax_origin.add_patch(c2)
    ax_origin.add_patch(p)

    return c1, c2, bbox_zoomed_patch, bbox_origin_patch, p

def zoom_effect02(ax1, ax2, **kwargs):
    """
    ax1 : the main axes
    ax1 : the zoomed axes

    Similar to zoom_effect01.  The xmin & xmax will be taken from the
    ax1.viewLim.
    """

    tt = ax1.transScale + (ax1.transLimits + ax2.transAxes)
    trans = blended_transform_factory(ax2.transData, tt)

    mybbox1 = ax1.bbox
    mybbox2 = TransformedBbox(ax1.viewLim, trans)

    prop_patches = kwargs.copy()
    prop_patches["ec"] = "none"
    prop_patches["alpha"] = 0.2


    c1, c2, bbox_patch1, bbox_patch2, p = \
        connect_bbox(mybbox1, mybbox2,
                     loc1a=2, loc2a=3, loc1b=1, loc2b=4,
                     prop_lines=kwargs, prop_patches=prop_patches)

    #ax1.add_patch(bbox_patch1)
    #ax2.add_patch(bbox_patch2)
    ax2.add_patch(c1)
    ax2.add_patch(c2)
    ax2.add_patch(p)

    return c1, c2, bbox_patch1, bbox_patch2, p




# Copyright 2015 Jeff Vierstra

from . import genomic_interval

def bed3_iterator(file_handle):
	"""
	Generator that parses BED3 format from a string iterator

	Returns:
		genomic_interval
	"""

	for line in file_handle:
		
		fields = line.strip().split()
		
		chrom = fields[0]
		start = int(fields[1])
		end = int(fields[2])

		yield genomic_interval(chrom, start, end)

def bed5_iterator(file_handle):
	"""
	Generator that parses BED5 format from a string iterator
	
	Returns:
		genomic_interval
	"""

	for line in file_handle:
		
		fields = line.strip().split()
		
		chrom = fields[0]
		start = int(fields[1])
		end = int(fields[2])
		name = fields[3]
		score = float(fields[4])

		yield genomic_interval(chrom, start, end, name, score)


def bed6_iterator(file_handle):
	"""
	Generator that parses BED6 format from a string iterator
	
	Returns:
		genomic_interval
	"""

	for line in file_handle:
		
		fields = line.strip().split()
		
		chrom = fields[0]
		start = int(fields[1])
		end = int(fields[2])
		name = fields[3]
		score = float(fields[4])
		strand = fields[5]

		yield genomic_interval(chrom, start, end, name, score, strand)



import numpy as numpy
import pysam


from .. import load_segmented_data

from matplotlib.patches import Rectangle
from matplotlib.transforms import Affine2D

from .track import track, pack_rows


class segment_track(track):

	def __init__(self, interval, **kwargs):
		super(segment_track, self).__init__(interval, **kwargs)

		self.segments = []

	def load_data(self, filepath, filterfunc=None):
		segments = load_segmented_data(filepath, self.interval)
		self.segments = filter(filterfunc, segments) if filterfunc else segments

	def format_axis(self, ax):
		super(segment_track, self).format_axis(ax)

		ax.xaxis.set(visible = False)
		ax.yaxis.set(visible = False)

	def render(self, ax, pack = True, labels = False):

		self.format_axis(ax)

		nrows, rows = pack_rows(self.segments)
		
		for segment in self.segments:

			n = rows[segment] if pack else 1
							
			x0 = segment.start
			x1 = segment.end
			y0 = -0.2+n
			w = x1-x0
			h = 0.4

			label_x, label_y, label_rot = x1, y0, 0

			p = Rectangle((x0, y0), w, h, 
				edgecolor=self.options["edgecolor"], 
				facecolor=self.options["facecolor"],
				lw=self.options["lw"],
				ls=self.options["ls"],
				zorder=1)

			ax.add_patch(p)
			
			if labels:
				ax.text(label_x, label_y, segment.name, rotation=label_rot, ha='left', va='bottom')

		ax.set_ylim(bottom=0.5, top=(nrows if pack else 1) + 1.5)

		super(segment_track, self).render(ax)



__all__ = ["track", "continuous_data_track", "gencode_annotation_track", "segment_track", "heatmap_track", "ideogram", "pwm", "connectors"]

from .ideogram import ideogram
from .pwm import pwm

from .track import track
from .continuous_data_track import continuous_data_track
from .gencode_annotation_track import gencode_annotation_track
from .segment_track import segment_track
from .heatmap_track import heatmap_track


import numpy as np
import pysam

from matplotlib.patches import Rectangle

from .. import genomic_interval
from .track import track, pack_rows

class gencode_annotation_track(track):

	def __init__(self, interval,  **kwargs):
		super(gencode_annotation_track, self).__init__(interval, **kwargs)

		self.genes = {}
		self.transcripts = {}
		self.exons = {}
		self.cds = {}
		self.utrs = {}

	def parse_gff_attributes(self, s):
		attr = {}
		for field in s.split(';'):
			(key, val) = field.split('=')
			attr[key] = val
		return attr

	def load_data(self, filepath):
		"""Loads GFF data"""
		tabix = pysam.TabixFile(filepath)

		for row in tabix.fetch(self.interval.chrom, self.interval.start, self.interval.end, parser = pysam.asGTF()):

			feature_interval = genomic_interval(row.contig, row.start, row.end, strand = row.strand)
			a = self.parse_gff_attributes(row.attributes)

			if row.feature == "gene":
				self.genes[a["gene_id"]] = a["gene_name"]
			elif row.feature == "transcript":
				self.transcripts[a["gene_id"]] = self.transcripts.get(a["gene_id"], []) + [(a["transcript_id"], feature_interval)]
			elif row.feature == "exon":
				self.exons[a["transcript_id"]] = self.exons.get(a["transcript_id"], []) + [(a["exon_id"], feature_interval)]
			elif row.feature == "CDS":
				self.cds[a["exon_id"]] = self.cds.get(a["exon_id"], []) + [(a["ID"], feature_interval)]
			elif row.feature == "UTR":
				self.utrs[a["exon_id"]] = self.utrs.get(a["exon_id"], []) + [(a["ID"], feature_interval)]

		tabix.close()
	
	def get_transcript_intervals(self):
		transcript_intervals = []
		for gene, trans in self.transcripts.items():
			for transcript_name, transcript_interval in trans:
				transcript_intervals.append(transcript_interval)

		return transcript_intervals

	def format_axis(self, ax):

		super(gencode_annotation_track, self).format_axis(ax)

		ax.xaxis.set(visible = False)
		ax.yaxis.set(visible = False)
		
	def render(self, ax, padding = 5):

		self.format_axis(ax)
		self.format_spines(ax, remove_spines=['top', 'bottom', 'left', 'right'])
		# pack transcripts

		trans = self.get_transcript_intervals()
		nrows, rows = pack_rows(trans, padding = padding)

		label_xoffset = len(self.interval) * 0.04

		label_arrow_height = 0.3
		label_arrow_width = len(self.interval) * 0.025

		label_arrow_xoffset = 0
		label_arrow_yoffset = 0.2

		for gene, gene_name in self.genes.items():
	
			if not gene in self.transcripts:
				continue
	
			for (transcript, transcript_interval) in self.transcripts[gene]:
			
				n = rows[transcript_interval]

				x0 = transcript_interval.start #- self.interval.start
				x1 = transcript_interval.end  #- transcript_interval.start + x0
				y = n

				over_left = x0 < self.interval.start #0
				over_right = x1 > self.interval.end #len(self.interval)
				start_visible = False
			
				ax.plot([max(x0, self.interval.start), min(x1, self.interval.end)], [y, y], color = 'grey', lw = 2, zorder = 0)

				if transcript_interval.strand == '+':

					label_arrow_xcoords = np.array([x0, x0, x0 + label_arrow_width])
					label_arrow_ycoords = np.array([y, y + label_arrow_height, y + label_arrow_height ]) + label_arrow_yoffset
					ma = '>'

					if not over_left:

						start_visible = True

						label_xcoord = x0 + label_xoffset
						label_ycoord = y + label_arrow_height + label_arrow_yoffset
						ha = "left"

					else:

						label_xcoord = self.interval.start - label_xoffset
						label_ycoord = y
						ha = "right"

				else: 

					label_arrow_xcoords = np.array([x1, x1, x1 - label_arrow_width])
					label_arrow_ycoords = np.array([y, y + label_arrow_height, y + label_arrow_height ]) + label_arrow_yoffset
					ma = '<'

					if not over_right:

						start_visible = True

						label_xcoord = x1 - label_xoffset
						label_ycoord = y + label_arrow_height + label_arrow_yoffset
						ha = "right"

					else:

						label_xcoord = self.interval.end + label_xoffset
						label_ycoord = y
						ha = "left"

				if start_visible:
					ax.plot(label_arrow_xcoords, label_arrow_ycoords, color = 'black')
					ax.plot(label_arrow_xcoords[-1], label_arrow_ycoords[-1], marker = ma, color = 'black', markeredgecolor='none')

				ax.text(label_xcoord, label_ycoord, gene_name, style = 'italic', verticalalignment = 'center', horizontalalignment = ha)
				
				if over_left:
					ax.plot(self.interval.start, y, marker = ma, clip_on = False, color = 'grey', markeredgecolor='none')
				if over_right:
					ax.plot(self.interval.end, y, marker = ma, clip_on = False, color = 'grey', markeredgecolor='none')
					
				# Loop through exons
				for (exon, exon_interval) in self.exons.get(transcript, []):

					for (utr, utr_interval) in self.utrs.get(exon, []):
						p = Rectangle((utr_interval.start, -0.1+n), utr_interval.end-utr_interval.start, 0.2, edgecolor = 'none', facecolor = 'goldenrod', zorder = 1)
						ax.add_patch(p)

					for (cds, cds_interval) in self.cds.get(exon, []):
						p = Rectangle((cds_interval.start, -0.15+n), cds_interval.end-cds_interval.start, 0.3, edgecolor = 'none', facecolor = 'goldenrod', zorder = 2)
						ax.add_patch(p)
	
		ax.set_ylim(bottom = 0.5, top = nrows + 1.5)
		
		super(gencode_annotation_track, self).render(ax)
		



# Copyright 2016 Jeff Vierstra

import numpy as np

from .. import load_data
from .track import track

class heatmap_track(track):

	def __init__(self, interval, data, **kwargs):
		super(heatmap_track, self).__init__(interval, **kwargs)

		self.data = data if data.ndim>1 else data[np.newaxis,:]

	def load_data(self, filepath, column=5, dtype=float):
		self.data = load_data(filepath, interval, [column], dtype).T

	def format_axis(self, ax):
		super(heatmap_track, self).format_axis(ax)

		ax.yaxis.set(visible = True, ticks = np.arange(0.5, self.data.shape[0], 0.5), ticklabels = [])
		#ax.set_ylim([0, self.data.shape[0]])

	def render(self, ax):
		
		self.format_axis(ax)
		self.format_spines(ax, remove_spines=[])

		x = range(self.interval.start, self.interval.end+1)
		y = range(self.data.shape[0]+1)
		
		vmin = self.options.get('vmin', np.min(self.data))
		vmax = self.options.get('vmax', np.max(self.data))
		vmid = self.options.get('vmid', 0)

		if vmax>0 and vmin<0:
			divergent_data=True
			vmin+=vmid
			vmax+=vmid

		if not 'cmap' in self.options:
			if divergent_data:
				cmap='RdBu'
			else:
				cmap='RdPu'
		else:
			cmap = self.options['cmap']

		ax.pcolormesh(x, y, self.data[::-1,:], cmap = cmap, vmin=vmin, vmax=vmax, rasterized=True)

		super(heatmap_track, self).render(ax)


# Copyright 2019 Jeff Vierstra

# Draw an ideogram; one can download the ideogram files from the UCSC genome browser

from ..helpers import open_file


from matplotlib.collections import BrokenBarHCollection
from matplotlib.patches import Polygon

import numpy as np

color_lookup = {
                  'gneg': (1., 1., 1.),
                'gpos25': (.6, .6, .6),
                'gpos50': (.4, .4, .4),
                'gpos75': (.2, .2, .2),
               'gpos100': (0., 0., 0.),
                  'acen': (.8, .4, .4),
                  'gvar': (.8, .8, .8),
                 'stalk': (.9, .9, .9),
               }

class ideogram(object):

    def __init__(self, filepath="/home/jvierstra/data/genomes/hg38/cytoBandIdeo.txt"):
        self.xranges = None
        self.colors = None
        self.centromeres = None

        self.read_ideogram(filepath)

    def read_ideogram(self, filepath):
        
        xranges={}
        colors={}
        centromeres={}

        f = open_file(filepath)

        last_chrom=None
        xr=[]
        for line in f:
            chrom, start, stop, label, stain = line.strip().split('\t')
            start = int(start)
            stop = int(stop)
            width = stop - start
        
            if stain=="acen":
                centromeres[chrom] = centromeres.get(chrom, []) + [(start, width)]
                continue
        
            xranges[chrom] = xranges.get(chrom, []) + [(start, width)]
            colors[chrom] = colors.get(chrom, []) + [(color_lookup[stain])]

        f.close()

        self.xranges = xranges
        self.colors = colors
        self.centromeres = centromeres

    def render(self, ax, chrom, pos = None):
        
        try:
            xranges = self.xranges[chrom]
            colors = self.colors[chrom]
            centromeres = self.centromeres[chrom]
        except:
            print("Error: No chromosome named: {}".format(chrom))
            return

        yranges = (0, 0.5)
        
        coll = BrokenBarHCollection(xranges, yranges, facecolors=colors, edgecolors='black', linewidths=0.5)    
        ax.add_collection(coll)

        if pos:
            ax.axvline(pos, color='red', lw=4)
        w = xranges[-1][0]+xranges[-1][1]
        
        pad=w*0.05
        
        ax.set_xlim(0-pad, xranges[-1][0]+xranges[-1][1]+pad)
        ax.xaxis.set_visible(False)
        
        center = yranges[0] + yranges[1]/2.
        
     
        x0, y0 = centromeres[0][0], yranges[0]
        x1, y1 = centromeres[0][0], yranges[1]
        x2, y2 = centromeres[0][0]+centromeres[0][1], center
        cent = Polygon(np.array([[x0, y0], [x1, y1], [x2, y2]]), closed=True, 
                       fc=color_lookup['acen'], ec='black', linewidth=0.5)
        ax.add_patch(cent)
        
        x0, y0 = centromeres[1][0], center
        x1, y1 = centromeres[1][0]+centromeres[1][1], yranges[1]
        x2, y2 = centromeres[1][0]+centromeres[1][1], yranges[0]
        
        cent = Polygon(np.array([[x0, y0], [x1, y1], [x2, y2]]), closed=True,
                        fc=color_lookup['acen'], ec='black', linewidth=0.5)
        ax.add_patch(cent)
        
        
        ax.set_yticks([center])
        ax.set_yticklabels([chrom])
        ax.set_ylim(-0.2, 0.7)
        
        [ax.spines[loc].set_color('none') for loc in ['top', 'left','right','bottom']]


# Copyright 2019 Jeff Vierstra

import gzip

magic_dict = {
    b"\x1f\x8b\x08": "gz",
    b"\x42\x5a\x68": "bz2",
    b"\x50\x4b\x03\x04": "zip"
}

max_len = max(len(x) for x in magic_dict)

def get_file_type(filename):
    with open(filename, 'rb') as f:
        file_start = f.read(max_len)
        for magic, filetype in magic_dict.items():
            if file_start.startswith(magic):
                return filetype
        return None

def open_file(filename):
	file_type = get_file_type(filename)
	if file_type == "gz":
		return gzip.open(filename, mode='rt')
	else:
		return open(filename)


from __future__ import absolute_import, division, print_function

import os
import sys

from setuptools import find_packages, setup

__version__ = "1.0.1"

install_requires = ["numpy>1.10", "scipy", "pysam>=0.8.2", "matplotlib"] 

setup(
	name = "genome_tools",
	version = __version__,
	license = "GPL-3.0-or-later",
	description = "A toolkit for processing and plotting genomic datasets.",
	long_description = "",
	author = "Jeff Vierstra",
	author_email = "jvierstra@altius.org",
	zip_safe = False,
	packages =  find_packages(),
    install_requires = install_requires,
    download_url = "https://github.com/jvierstra/genome-tools/archive/v1.0.1.tar.gz",
    classifiers=[
	    'Development Status :: 5 - Production/Stable',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
	    'Intended Audience :: Developers',      # Define that your audience are developers
	    'Topic :: Software Development :: Build Tools',
	    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
	    'Programming Language :: Python :: 2.7',     #Specify which pyhton versions that you want to support
	    'Programming Language :: Python :: 3',
	    'Programming Language :: Python :: 3.5', 
    ],
)



from matplotlib.patheffects import RendererBase
from matplotlib.transforms import offset_copy
from matplotlib.patches import Rectangle

import numpy as np

import seaborn as sns

class Scale(RendererBase):
    def __init__(self, sx, sy=None):
        self._sx = sx
        self._sy = sy
        
    def draw_path(self, renderer, gc, tpath, affine, rgbFace):
        affine = affine.identity().scale(self._sx, self._sy) + affine
        renderer.draw_path(gc, tpath, affine, rgbFace)
        
basic = "ACGT"
iupac = "XACMGRSVTWYHKDBN"

bg_uni =  np.array([0.25, 0.25, 0.25, 0.25])
bg = np.array([0.2, 0.3, 0.3, 0.2])

iupac_colors = ['black'] * 16
iupac_colors[1<<0] = 'green'
iupac_colors[1<<1] = 'blue'
iupac_colors[1<<2] = 'gold'
iupac_colors[1<<3] = 'red'


class pwm(object):

    def __init__(self, mat):
        self.data = mat

    def rel_info_content(self, bg=None):
        mat = np.copy(self.data).T
        p=mat/np.sum(mat, axis = 1)[:,np.newaxis]
        if bg is None:
            ic=2+np.sum(p*np.nan_to_num(np.log2(p)), axis = 1)
        else:
            ic=2+np.sum(p*np.nan_to_num(np.log2(p/bg[np.newaxis,:])), axis = 1)
            
        ric=p*ic[:,np.newaxis]

        return ric

    def seq_logp(self, seq, bg):
        res=0
        for i, c in enumerate(seq):
            j = basic.find(c)
            res += np.log(self.data[i, j]/bg[j])
        return res

    def render(self, fig, ax, pad=None, xoffset=0, xlim=None, type='default', bg=None, rc=False):

        if type=='ic':
            mat=self.rel_info_content(bg=bg).T
            ax.set_ylim(0, 2.1)
        elif type=='affinity':
            ddg = np.log(self.data+1e-3)
            mat=ddg-np.mean(ddg, axis=0)[np.newaxis,:]
            top=np.max(np.ma.sum(np.ma.array(mat,  mask=mat<0), axis=0))
            bottom=np.min(np.ma.sum(np.ma.array(mat,  mask=mat>=0), axis=0))
            ax.set_ylim(bottom, top)
        else:
            mat=self.data
            ax.set_ylim(0, 1.05)
        
        if rc:
            mat = mat[::-1,:][:,::-1]
            

        w = mat.shape[1]
        if xlim:
            ax.set_xlim(xlim[0], xlim[1])
        else:
            ax.set_xlim(xoffset-pad, xoffset+w+pad)
        
        wscale=1
        init = False

        for i in range(w):

            heights=mat[:,i]
            order=np.argsort(heights)
            reorder=np.concatenate([order[heights[order]>0], order[heights[order]<=0]])

            yshift_pos = 0
            yshift_neg = np.sum(heights[heights<=0])

            for j, b in enumerate(reorder):
                base=iupac[(1<<b)]
                color=iupac_colors[(1<<b)]
                scale=heights[b]

                if scale>0:
                    yshift=yshift_pos
                    alpha=1
                else:
                    yshift=yshift_neg
                    alpha=0.4

                h=np.abs(scale)

                t = ax.text(i+xoffset, yshift, base, ha='left', va='baseline', 
                    color=color, fontsize=40, family='monospace', 
                    fontname="Bitstream Vera Sans Mono", weight='bold', alpha=alpha)

                if not init:
                    fig.canvas.draw()
                    ext = t.get_window_extent(t._renderer)

                    x0, x1, y0, y1 = ext.x0, ext.x1, ext.y0, ext.y1
                    nx0, ny0 = ax.transData.inverted().transform((x0, y0))
                    nx1, ny1 = ax.transData.inverted().transform((x1, y1))
                    
                    wscale=1.0/(nx1-nx0)
                    hscale=(1.0/(ny1))
                    
                    height=ext.height        
                    
                    init=True

                t.set_path_effects([Scale(wscale, hscale*h)])
                
                if scale>0:
                    yshift_pos+=h
                else:
                    yshift_neg+=h
                
        ax.set_xticks(np.arange(xoffset, xoffset+mat.shape[1])+0.5)
        ax.set_xticklabels([])
        ax.xaxis.set_visible(False)
        
        ax.set_yticks([])
        ax.set_yticklabels([])
        #ax.yaxis.set_visible(False)
        
        sns.despine(ax=ax, left=True, bottom=True)

    


# Copyright 2016 Jeff Vierstra

import sys

from matplotlib.ticker import MaxNLocator, FuncFormatter
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar



class track(object):
    """Base level track object"""
    def __init__(self, interval, **kwargs):
        self.interval = interval
        self.orientation = 'horizontal'
        self.options = {
            'facecolor': 'b',
            'edgecolor': 'none',
            'lw': 1,
            'ls': '-',
            'scale_bar': None,
            'scale_bar_loc': 1
        }

        self.options.update(kwargs)

    def format_axis(self, ax):

        # set defaults
        ax.ticklabel_format(style='plain', axis = 'both', useOffset=False)
                
        # tick direction out
        ax.xaxis.set_tick_params(direction = 'out')
        ax.yaxis.set_tick_params(direction = 'out')
        
        # set x limits to interval
        ax.set_xlim(left=self.interval.start, right=self.interval.end)
        
        # x-axis choose ~3 tick positions
        locator = MaxNLocator(3, prune = 'both')
        ax.xaxis.set(major_locator = locator)
        
        # hide both x and y axis
        ax.xaxis.set(visible = False)
        ax.yaxis.set(visible = False)


    def format_spines(self, ax, remove_spines):
        all_spines = ['top', 'bottom', 'left', 'right']
        for spine in remove_spines:
            try:
                ax.spines[spine].set_visible(False)
            except KeyError:
                pass

        for spine in set(all_spines).difference(set(remove_spines)):
            try:
                ax.spines[spine].set_linewidth(0.5)
            except:
                pass

    def render(self, ax):
        # Remove the white patch behind each axes
        ax.patch.set_facecolor('none')

        # Add scale bar -- code is placed here so a scale bar can be drawn on any track instance
        if self.options['scale_bar'] is not None:
            bar = AnchoredSizeBar(ax.transData,
                self.options['scale_bar'], 
                label="%d nt" % self.options['scale_bar'], 
                loc=self.options['scale_bar_loc'],
                frameon=False)
            ax.add_artist(bar)

    def load_data(self, filepath):
        pass

class row_element(object):

    def __init__(self, interval, prev = None, next = None, padding = 5):
        self.interval = interval
        self.start = interval.start - padding
        self.end = interval.end + padding
        self.prev = prev
        self.next = next
        
class row(object):
    """ """
    def __init__(self, num):
        self.num = num
        self.elements = []
        self.first = None
        self.last = None
        
    def add_element(self, elem):
        
        if self.first is None:
            #sys.stderr.write("added feat %d-%d as only element of row %d\n" %
            #                (elem.start, elem.end, self.num))
            elem.prev = None
            elem.next = None
            self.first = self.last = elem
            return True
        
        cur = self.first
        while (cur and cur.start < elem.start):
            cur = cur.next
            
        if cur is None:
            if self.last.end < elem.start:
                #sys.stderr.write("added feat %d-%d to end of row %d\n" %
                #                 (elem.start, elem.end, self.num))
                elem.prev = self.last
                elem.next = None
                self.last.next = elem
                self.last = elem
                return True
            else:
                return False
        
        prev = cur.prev
        if prev is None:
            if elem.end < cur.start:
                #sys.stderr.write("added feat %d-%d to start of row %d "
                #                 "(before %d-%d)\n" %
                #                 (elem.start, elem.end, self.num,
                #                 cur.start, cur.end))
                elem.prev = None
                elem.next = cur
                cur.prev = elem
                self.first = elem
                return True
            else:
                return False
        
        if prev.end < elem.start and cur.start > elem.end:
            #sys.stderr.write("added feat %d-%d between %d-%d and %d-%d "
            #                 "in row %d\n" %
            #                 (elem.start, elem.end,
            #                  prev.start, prev.end,
            #                  cur.start, cur.end, self.num))
            elem.prev = prev
            elem.next = cur
            prev.next = elem
            cur.prev = elem
            return True
        
        return False

def pack_rows(intervals, padding = 5):
    
    rows = []
    row_num = {}
    
    nrows = 0

    for interval in intervals:
        #sys.stderr.write("adding feat...\n")
        re = row_element(interval, padding = padding)
        
        placed = False
        
        for r in rows:
            if(r.add_element(re)):
                placed = True
                row_num[interval] = r.num
                break
        
        if not placed:
            nrows = nrows + 1
            r = row(num = nrows)
            rows.append(r)
            row_num[interval] = r.num
            if not r.add_element(re):
                raise Exception("Could not place element in new row!")
                
    return nrows, row_num

def segment(x, threshold, w = 0, decreasing = False):
    """Segment an array into continuous elements passing a threshhold

    :returns: [(int, int), ...]: start and end points to regions that pass a threshold
    """
    dir = -1 if decreasing else 1

    ret = []
    curr_start = -1
    
    for i in range(x.shape[0]):
        if curr_start < 0:
            if dir*x[i] >= dir*threshold:
                curr_start = i-w+1
        else:
            if dir*x[i] < dir*threshold:
                if len(ret) > 0 and curr_start <= ret[-1][1]:
                    ret[-1][1] = i-1+w
                else:
                    ret.append( [curr_start, i-1+w] )
                curr_start = -1
    return ret
